// SleepTest.cpp : Defines the entry point for the console application.
//

#include <Windows.h>
#include <stdio.h>
#include <tchar.h>
#include <process.h>

//////////////////////////////////////////////////////////////////////////
// Just some tuning parameters

const int iNumThreads = 3;			// Controls number of threads to create
const long lMax = 10000000;			// Controls the outer loop 
const long lMaxInnerLock = 50;		// Controls duration of the lock

// Global data for the threads
volatile long lVariable;	// Just a counter we use 
volatile long lDummy;		// Just a global variable to do some silly work
volatile long lLock;		// Lock variable for the spin lock

// Modes to use to wait and spin for the lock
enum eMode
{
	modeNoop,
	modeSleep0,
	modeSwitchToThread,
	modeSleep1,
} mode;

unsigned __stdcall ThreadFunc( void * ptr )
{
	bool bStop = false;
	long *plCount = static_cast<long*>(ptr);
	do 
	{
		// Do a spin lock
		while (InterlockedExchange(&lLock,1)!=0)
		{
			switch (mode)
			{
			case modeNoop:
				__noop;
				break;
			case modeSleep0:
				Sleep(0);
				break;
			case modeSwitchToThread:
				SwitchToThread();
				break;
			case modeSleep1:
				Sleep(1);
				break;
			}
		}

		// Check if we are done
		if (lVariable==lMax)
			bStop = true;
		else
		{
			++lVariable;
			++*plCount;
		}

		// Just delay the operation a little bit.
		// Not that this loop isn't optimized and does not disappear
		// though code optimizing!
		for (int i=0; i<lMaxInnerLock; ++i)
			// Do some silly operation
			lDummy += GetTickCount();

		// Unlock
		InterlockedExchange(&lLock,0);
	} 
	while (!bStop);
	return 0;
}

void Test()
{
	DWORD dwStart = GetTickCount();
	lVariable = 0;
	long lCount[iNumThreads] = { 0 };
	HANDLE hThread[iNumThreads] = { NULL };
	for (int i=0; i<_countof(hThread); ++i)
		hThread[i] = reinterpret_cast<HANDLE>(_beginthreadex(0,0,ThreadFunc,&lCount[i],CREATE_SUSPENDED,NULL));

	// 1. Thread highest prio
	// 2. Thread above normal prio
	// 3. Thread normal prio
	// 4. Thread below normal prio
	// 5. Thread lowest prio
	SetThreadPriority(hThread[0],THREAD_PRIORITY_HIGHEST);
	if (iNumThreads>1)
		SetThreadPriority(hThread[1],THREAD_PRIORITY_ABOVE_NORMAL);
	if (iNumThreads>2)
		SetThreadPriority(hThread[2],THREAD_PRIORITY_NORMAL);
	if (iNumThreads>3)
		SetThreadPriority(hThread[3],THREAD_PRIORITY_BELOW_NORMAL);
	if (iNumThreads>4)
		SetThreadPriority(hThread[4],THREAD_PRIORITY_LOWEST);

	for (int i=0; i<_countof(hThread); ++i)
		ResumeThread(hThread[i]);

	for (int i=0; i<_countof(hThread); ++i)
		WaitForSingleObject(hThread[i],INFINITE);
	
	for (int i=0; i<_countof(hThread); ++i)
		CloseHandle(hThread[i]);

	DWORD dwEnd = GetTickCount();
	for (int i=0; i<_countof(hThread); ++i)
		printf("Count %d:  %8d\n", i+1, lCount[i]);
	printf("Duration: %8d\n", dwEnd-dwStart);
}

int _tmain(int argc, _TCHAR* argv[])
{
	// __noop;
	mode = modeNoop;
	printf("noop;\n");
	Test();

	// Sleep(0);
	mode = modeSleep0;
	printf("Sleep(0);\n");
	Test();

	// SwitchToThread();
	mode = modeSwitchToThread;
	printf("SwitchToThread();\n");
	Test();

	// Sleep(1);
	mode = modeSleep1;
	printf("Sleep(1);\n");
	Test();
}

